
source("code_hermite.r")
source("code_eta.r")
source("code_delta.r")
source("code_objective.r")
source("code_delaigle_gijbels.r")
library(doParallel);
cl<-makeCluster(13);
registerDoParallel(cl);
RESULTS <- foreach( iter = 1:2525, .combine = rbind )%dopar%{

#########################################################################
# Imposed Parameters 
#########################################################################
B = 1000           # Number of Multiplier Bootstrap
q = 6              # Sieve Dimension
L = 50             # Grid Size
alpha = 0.05       # Size

#########################################################################
# DGP Parameters
#########################################################################
n = 1000

df_X = 10
sd_X = df_X^0.5 / (df_X - 2)^0.5
sd_U1 = 0.5
sd_U2 = 0.5
x_list = -10:10/5 * sd_X  # Grid of X

#########################################################################
# DGP 
#########################################################################
X = c( rt( n, df_X ) )
f <- function(x){ dt( x, df_X ) }
U1 = rnorm( n, 0, sd_U1 )
U2 = rnorm( n, 0, sd_U2 )
Y1 = X + U1
Y2 = X + U2

#########################################################################
# Alternatives 
#########################################################################
list_loc = 1:25*0.04 + 0
list_sca = 1:25*0.04 + 1
list_df = 1:25 + df_X
f_location <- function(x,loc){ dt( x-loc, df_X ) }
f_scale <- function(x,sca){ dt( x/sca, df_X )/sca }
f_df <- function(x,df){ dt( x, df ) }

#########################################################################
# Data-Driven Tuning Parameters 
#########################################################################
eta = get_eta( q )
delta = get_delta( q, Y1 )
hstar = Delaigle_Gijbels(x_list,n,Y1,Y2)
T = 1/hstar

#########################################################################
# Construct L*(q+1) by n Matrices mat_R And mat_I 
#########################################################################
list_t = -T^(1/3)-2*T^(1/3)/(L-1) + 2*T^(1/3)/(L-1)*(1:L)
list_t = list_t^3
mat_R = NULL;
mat_I = NULL;
for( j in 0:q ){
 mat_cos_tY2 = cos( matrix( list_t, L, 1 ) %*% matrix( Y2, 1, n ) )
 mat_sin_tY2 = sin( matrix( list_t, L, 1 ) %*% matrix( Y2, 1, n ) )
 mat_Y1 = matrix( 1, L, 1 ) %*% matrix( Y1, 1, n )
 psij = fpsi( list_t, j )
 psij_prime = fpsi_prime( list_t, j )
 mat_Im_psij = Im( (0+1i)^j ) * matrix( psij, L, 1) %*% matrix( 1, 1, n )
 mat_Re_psij_prime = Re( (0+1i)^j ) * matrix( psij_prime, L, 1) %*% matrix( 1, 1, n )
 mat_Re_psij = Re( (0+1i)^j ) * matrix( psij, L, 1) %*% matrix( 1, 1, n )
 mat_Im_psij_prime = Im( (0+1i)^j ) * matrix( psij_prime, L, 1) %*% matrix( 1, 1, n )

 mat_Rj = (2*3.141593)^0.5 * ( -mat_cos_tY2 * ( mat_Y1*mat_Im_psij + mat_Re_psij_prime ) - mat_sin_tY2 * ( mat_Y1*mat_Re_psij - mat_Im_psij_prime ) )
 mat_Ij = (2*3.141593)^0.5 * (  mat_cos_tY2 * ( mat_Y1*mat_Re_psij + mat_Im_psij_prime ) - mat_sin_tY2 * ( mat_Y1*mat_Im_psij - mat_Re_psij_prime ) )

 mat_R = rbind( mat_R, mat_Rj )
 mat_I = rbind( mat_I, mat_Ij )
}

#########################################################################
# Construct L by (q+1) Matrices mat_E_R And mat_E_I
# Construct L*(q+1) by (q+1) Matrices mat_V_R And mat_V_I
#########################################################################
mat_E_R = NULL
mat_E_I = NULL
mat_V_R = NULL
mat_V_I = NULL
for( l in 1:L ){
 indices = 0:q*L + l
 mat_R_l = mat_R[indices,]
 mat_I_l = mat_I[indices,]
 E_R_l = rowMeans( mat_R_l )
 E_I_l = rowMeans( mat_I_l )
 V_R_l = mat_R_l %*% t(mat_R_l) / n - matrix( E_R_l, q+1, 1) %*% t(matrix( E_R_l, q+1, 1))
 V_I_l = mat_I_l %*% t(mat_I_l) / n - matrix( E_I_l, q+1, 1) %*% t(matrix( E_I_l, q+1, 1))

 mat_E_R = rbind( mat_E_R, E_R_l )
 mat_E_I = rbind( mat_E_I, E_I_l )
 mat_V_R = rbind( mat_V_R, V_R_l )
 mat_V_I = rbind( mat_V_I, V_I_l )
}

#########################################################################
# Construct L*B by (q+1) Matrices mat_E_R_MB And mat_E_I_MB
#########################################################################
mat_epsilon = matrix(rnorm(n*B),n,B)
mat_E_R_MB = NULL
mat_E_I_MB = NULL
for( l in 1:L ){
 indices = 0:q*L + l
 mat_E_R_MB_l = ( mat_R[indices,] - matrix( rowMeans(mat_R[indices,]), q+1, n) ) %*% mat_epsilon
 mat_E_I_MB_l = ( mat_I[indices,] - matrix( rowMeans(mat_I[indices,]), q+1, n) ) %*% mat_epsilon

 mat_E_R_MB = rbind( mat_E_R_MB, mat_E_R_MB_l )
 mat_E_I_MB = rbind( mat_E_I_MB, mat_E_I_MB_l )
}

#########################################################################
# Compute the Psi as length(x_list) by (q+1) Matrix psi_list
#########################################################################
psi_list = NULL
for( j in 0:q ){
 psi_list = cbind( psi_list, fpsi( -1*x_list, j ) )
}

#########################################################################
# Compute the Phi at 0 as 1 by (q+1) Matrix phi_list
#########################################################################
phi_list = NULL
for( j in 0:q ){
 phi_list = cbind( phi_list, fpsi( c(0,0), j ) )
}
phi_list = phi_list[1,]

#########################################################################
# Initial Value of Sieve Coefficients theta by Fourier Coefficients
#########################################################################
theta = matrix(0, q+1, 1)
for( j in 0:q ){ 
 theta[j+1] = sum( fpsi( -1*-1000:1000/100, j ) * f(-1000:1000/100) ) * 0.01
}

#########################################################################
# Penalization Parameter
#########################################################################
PPara = 10000
#########################################################################
# Constrained Optimization for Confidence Band
#########################################################################
Conf_Band = NULL
for( x_idx in 1:length(x_list) ){
 min_result = nlm( fobjective,theta,x_idx,FALSE,PPara,mat_E_R,mat_E_I,mat_E_R_MB,mat_E_I_MB,n,alpha,eta,delta,L,psi_list,phi_list)
 min_theta_estimate = min_result$estimate
 min_estimate = psi_list[x_idx,] %*% min_theta_estimate - eta
 max_result = nlm( fobjective,theta,x_idx,TRUE,PPara,mat_E_R,mat_E_I,mat_E_R_MB,mat_E_I_MB,n,alpha,eta,delta,L,psi_list,phi_list)
 max_theta_estimate = max_result$estimate
 max_estimate = psi_list[x_idx,] %*% max_theta_estimate + eta
 Conf_Band = rbind(Conf_Band, cbind(min_estimate,max_estimate))

 print(x_idx/length(x_list))
 Sys.sleep(0.001)
 flush.console()
}

for( x_idx in (1+2):(length(x_list)-2) ){
 if( Conf_Band[x_idx,1] < min(Conf_Band[x_idx-1,1],Conf_Band[x_idx+1,1]) ){
  min_result = nlm( fobjective,theta,x_idx-1,FALSE,PPara,mat_E_R,mat_E_I,mat_E_R_MB,mat_E_I_MB,n,alpha,eta,delta,L,psi_list,phi_list)
  min_theta_estimate_left = min_result$estimate
  min_result = nlm( fobjective,theta,x_idx+1,FALSE,PPara,mat_E_R,mat_E_I,mat_E_R_MB,mat_E_I_MB,n,alpha,eta,delta,L,psi_list,phi_list)
  min_theta_estimate_right = min_result$estimate
  min_result = nlm( fobjective,(min_theta_estimate_left+min_theta_estimate_right)/2,x_idx,FALSE,PPara,mat_E_R,mat_E_I,mat_E_R_MB,mat_E_I_MB,n,alpha,eta,delta,L,psi_list,phi_list)
  min_theta_estimate = min_result$estimate
  min_estimate = psi_list[x_idx,] %*% min_theta_estimate - eta
  Conf_Band[x_idx,1] = min_estimate
 }
 if( Conf_Band[x_idx,2] > min(Conf_Band[x_idx-1,2],Conf_Band[x_idx+1,2]) ){
  max_result = nlm( fobjective,theta,x_idx-1,TRUE,PPara,mat_E_R,mat_E_I,mat_E_R_MB,mat_E_I_MB,n,alpha,eta,delta,L,psi_list,phi_list)
  max_theta_estimate_left = max_result$estimate
  max_result = nlm( fobjective,theta,x_idx+1,TRUE,PPara,mat_E_R,mat_E_I,mat_E_R_MB,mat_E_I_MB,n,alpha,eta,delta,L,psi_list,phi_list)
  max_theta_estimate_right = max_result$estimate
  max_result = nlm( fobjective,(max_theta_estimate_left+max_theta_estimate_right)/2,x_idx,TRUE,PPara,mat_E_R,mat_E_I,mat_E_R_MB,mat_E_I_MB,n,alpha,eta,delta,L,psi_list,phi_list)
  max_theta_estimate = max_result$estimate
  max_estimate = psi_list[x_idx,] %*% max_theta_estimate + eta
  Conf_Band[x_idx,2] = max_estimate
 }
}

#########################################################################
# Summarize Results
#########################################################################
I = x_list
Conf_Band = Conf_Band
Cover = min( 1 * ( Conf_Band[,1] <= f(I) ) ) * min( 1 * ( f(I) <= Conf_Band[,2] ) ) == 1
Cover_Alt_Loc = NULL
Cover_Alt_Sca = NULL
Cover_Alt_df = NULL
for(idx in 1:25){
 Cover_Alt_Loc = cbind(Cover_Alt_Loc, min( 1 * ( Conf_Band[,1] <= f_location(I,list_loc[idx]) ) ) * min( 1 * ( f_location(I,list_loc[idx]) <= Conf_Band[,2] ) ) == 1)
 Cover_Alt_Sca = cbind(Cover_Alt_Sca, min( 1 * ( Conf_Band[,1] <= f_scale(I,list_sca[idx]) ) ) * min( 1 * ( f_scale(I,list_sca[idx]) <= Conf_Band[,2] ) ) == 1)
 Cover_Alt_df = cbind(Cover_Alt_df, min( 1 * ( Conf_Band[,1] <= f_df(I,list_df[idx]) ) ) * min( 1 * ( f_df(I,list_df[idx]) <= Conf_Band[,2] ) ) == 1)
}
Avg_Length = mean( Conf_Band[,2] - Conf_Band[,1] )
Med_Length = median( Conf_Band[,2] - Conf_Band[,1] )

#########################################################################
# Display Output
#########################################################################
etaX2 = 2*eta
Avg_Length_etaX2 = Avg_Length - etaX2
Med_Length_etaX2 = Med_Length - etaX2
filename = paste("RESULTS_T_dfX",df_X,"_sdU1_",100*sd_U1,"_sdU2_",100*sd_U2,"_n",n,"_q",q,"_L",L,".csv",sep="")
write.table(cbind( Cover, Avg_Length, Med_Length, etaX2, Avg_Length - etaX2, Med_Length - etaX2, Cover, Cover_Alt_Loc, Cover, Cover_Alt_Sca, Cover, Cover_Alt_df), file = filename, row.names = FALSE, col.names = FALSE,append=TRUE, sep = ",")
cbind( Cover, Avg_Length, Med_Length, etaX2, Avg_Length - etaX2, Med_Length - etaX2)
}
stopCluster(cl);
RESULTS
